function eventEmitter(valid_events) {
	let methods = {};
	let registered_handler = {};
	let event_handler_added_handler = null;
	let event_dispatched_handler = null;
	let event_handler_removed_handler = null;

	for (let event of valid_events) {
		registered_handler[event] = [];
	}

	const handlerInstalledForEvent = (event_name, fn) => {
		return registered_handler[event_name].filter(handler => {
			return Object.is(handler, fn)
		}).length > 0
	};

	methods.on = function on(event_name, fn) {
		return methods.addEventListener(event_name, fn)
	};

	methods.addEventListener = function addEventListener(event_name, fn) {
		if (!(event_name in registered_handler)) {
			throw new Error(`Invalid event '${event_name}'.`)
		} else if (handlerInstalledForEvent(event_name, fn)) {
			throw new Error(`This *exact* function was already registered for this event.`)
		}

		registered_handler[event_name].push(fn);

		if (typeof event_handler_added_handler === "function") {
			event_handler_added_handler(event_name, fn);
		}
	};

	methods.removeEventListener = function removeEventListener(event_name, fn) {
		if (!(event_name in registered_handler)) {
			throw new Error(`Invalid event '${event_name}'.`)
		} else if (!handlerInstalledForEvent(event_name, fn)) {
			throw new Error(`This function was not registered for this event.`)
		}

		registered_handler[event_name] = registered_handler[event_name].filter(handler => {
			return !Object.is(handler, fn)
		});

		if (typeof event_handler_removed_handler === "function") {
			event_handler_removed_handler(event_name, fn);
		}
	};

	return {
		setOnEventHandlerAddedHandler(fn) {
			event_handler_added_handler = fn;
		},

		setOnEventDispatchedHandler(fn) {
			event_dispatched_handler = fn;
		},

		setOnEventHandlerRemovedHandler(fn) {
			event_handler_removed_handler = fn;
		},

		install(obj) {
			for (const method of ["on", "addEventListener", "removeEventListener"]) {
				obj[method] = methods[method];
			}

			return function dispatchEvent(event_name, ...event_data) {
				if (!(event_name in registered_handler)) {
					throw new Error(`Invalid event '${event_name}'.`)
				}

				//
				// todo: maybe add "blocking" events
				// i.e. each event handler gets executed consecutively
				// example of code flow:
				// await handler()
				// await next_handler()
				// await all_handlers_dispatched()
				//
				for (const handler of registered_handler[event_name]) {
					setTimeout(handler, 0, ...event_data);
				}

				if (typeof event_dispatched_handler === "function") {
					setTimeout(event_dispatched_handler, 0, event_name, event_data);
				}
			}
		}
	}
}

function createPromise() {
	let promise, resolve, reject;

	promise = new Promise((a, b) => {
		resolve = a;
		reject = b;
	});

	return {promise, resolve, reject}
}

function createRandomIdentifier(len) {
	let str = "";
	let bytes = [];

	if (typeof crypto === "object" && typeof crypto.getRandomValues === "function") {
		bytes = new Uint8Array(len);

		crypto.getRandomValues(bytes);
	} else {
		for (let i = 0; i < len; ++i) {
			// will never exceed 255
			bytes.push(Math.floor(Math.random() * 256));
		}
	}

	for (let i = 0; i < len; ++i) {
		str += (bytes[i]).toString(16).slice(0, 1);
	}

	// sanity check output
	if (str.length !== len) {
		throw new Error(`createRandomIdentifier(${len}) failed for unknown reasons.`)
	}

	return str
}

function setupClient(id) {
	let client_instance = {
		id,
		current_stream: null,
		message_queue: [],
		public_interface: {
			id,
			sendMessage(msg) {
				// client is connected, send immediately
				if (client_instance.current_stream !== null) {
					client_instance.current_stream.write(
						`${JSON.stringify({
							message: msg
						})}\n`
					);
				} else {
					client_instance.message_queue.push(msg);
				}
			}
		}
	};

	const event_emitter = eventEmitter(["message", "disconnect"]);

	client_instance.dispatchEvent = event_emitter.install(client_instance.public_interface);

	return client_instance
}

function handleRequest(instance, relative_path, request, response) {
	if (relative_path === "create") {
		const client_id = createRandomIdentifier(32);

		instance.connected_clients.set(client_id, setupClient(client_id));

		response.write(JSON.stringify({
			id: client_id
		}) + "\n");
		response.end();

		return
	}
	else if (relative_path.startsWith("stream/")) {
		let client_id = relative_path.slice("stream/".length);

		if (!instance.connected_clients.has(client_id)) {
			throw new Error(`Invalid client identifier '${client_id}'.`)
		}

		const client_instance = instance.connected_clients.get(client_id);

		if (client_instance.current_stream !== null) {
			throw new Error(`Concurrent streams are not supported.`)
		}

		// release stream when socket closes
		request.socket.on("close", () => {
			client_instance.current_stream = null;

			instance.dispatchEvent("disconnect", client_instance.id);

			client_instance.dispatchEvent("disconnect");

			instance.connected_clients.delete(client_instance.id);
		});

		client_instance.current_stream = response;

		response.writeHead(200, {
			//"Content-Type": "text/event-stream",
			"Cache-Control": "no-cache",
			"Connection": "keep-alive"
		});

		instance.dispatchEvent("connect", client_instance.public_interface);

		client_instance.current_stream.write(
			`C813A843-5765-4DB3-8439-24AC1849117D\n`
		);

		// dump accumulated messages
		while (client_instance.message_queue.length) {
			const msg = client_instance.message_queue.shift();

			client_instance.current_stream.write(
				`${JSON.stringify({message: msg})}\n`
			);
		}

		return
	}
	else if (relative_path.startsWith("sendMessage/")) {
		let client_id = relative_path.slice("sendMessage/".length);

		if (!instance.connected_clients.has(client_id)) {
			throw new Error(`Invalid client identifier '${client_id}'.`)
		}

		const client_instance = instance.connected_clients.get(client_id);

		let body = "";

		request.on("data", chunk => body += chunk);

		request.on("end", () => {
			response.write("ok");
			response.end();

			const data = JSON.parse(body);

			instance.dispatchEvent("message", client_instance.id, data.message);
			client_instance.dispatchEvent("message", data.message);
		});

		return
	}

	throw new Error(`Unknown method '${relative_path}'.`)
}

async function index$1(http_port, base_url) {
	const {default: path} = await import('node:path');
	const {default: http} = await import('node:http');

	let instance = {
		ready_promise: createPromise(),
		connected_clients: new Map(),
		public_interface: {
			sendMessage(client_id, msg) {
				if (!instance.connected_clients.has(client_id)) {
					throw new Error(`No such client '${client_id}'.`)
				}

				const client_instance = instance.connected_clients.get(client_id);

				return client_instance.public_interface.sendMessage(msg)
			}
		}
	};

	const event_emitter = eventEmitter(["httpRequest", "connect", "message", "disconnect"]);

	instance.dispatchEvent = event_emitter.install(instance.public_interface);

	const normalized_base_url = path.normalize(base_url + "/");

	const server = http.createServer(async (request, response) => {
		const normalized_path = path.normalize(request.url);

		if (!normalized_path.startsWith(normalized_base_url)) {
			instance.dispatchEvent("httpRequest", {request, response});
		} else {
			const relative_path = normalized_path.slice(normalized_base_url.length);

			try {
				await handleRequest(instance, relative_path, request, response);
			} catch (error) {
				response.writeHead(500);
				response.write(`${error.message}`);
				response.end();
			}
		}
	});

	server.on("error", instance.ready_promise.reject);

	server.listen(http_port, () => {
		instance.public_interface.port = server.address().port;

		instance.ready_promise.resolve(instance.public_interface);
	});

	return instance.ready_promise.promise
}

//
// code taken from https://rob-blackbourn.medium.com/beyond-eventsource-streaming-fetch-with-readablestream-5765c7de21a1
// and adjusted to fit our use case
//
// Original Author: Rob Blackbourn, Date: Sep 23, 2019
//


function createLineDecoder() {
	return new TransformStream({
		start(controller) {
			controller.buf = "";
			controller.pos = 0;
		},

		transform(chunk, controller) {
			controller.buf += chunk;

			while (controller.pos < controller.buf.length) {
				if (controller.buf[controller.pos] == "\n") {
					const line = controller.buf.substring(0, controller.pos);
					controller.enqueue(line);
					controller.buf = controller.buf.substring(controller.pos + 1);
					controller.pos = 0;
				} else {
					controller.pos++;
				}
			}
		}
	})
}

function createWriteableEventStream(dispatchEvent) {
	return new WritableStream({
		start(controller) {
			dispatchEvent("start");
		},

		write(message, controller) {
			dispatchEvent("message", message);
		},

		close(controller) {
			dispatchEvent("close");
		},

		abort(reason) {
			dispatchEvent("close");
		}
	})
}

function SSEWithFetch(url, options = {}) {
	let instance = {
		public_interface: {}
	};

	const event_emitter = eventEmitter(["start", "message", "close", "error"]);

	const dispatchEvent = event_emitter.install(instance.public_interface);

	const line_decoder = createLineDecoder();
	const event_stream = createWriteableEventStream(dispatchEvent);

	fetch(url, options)
	.then(response => {
		return response.body
			.pipeThrough(new TextDecoderStream())
			.pipeThrough(line_decoder)
			.pipeTo(event_stream)
	})
	.catch(error => {
		dispatchEvent("error", error);
	});

	return instance.public_interface
}

async function createClient(endpoint_url) {
	const response = await fetch(`${endpoint_url}/create`);

	return (await response.json()).id
}

async function index(endpoint_url) {
	const ready_promise = createPromise();

	const id = await createClient(endpoint_url);

	let instance = {
		id,
		is_ready: false,
		message_queue: [],
		public_interface: {
			id,

			async sendMessage(msg) {
				await fetch(`${endpoint_url}/sendMessage/${id}`, {
					headers: {
						"Content-Type": "application/json"
					},
					method: "POST",
					body: JSON.stringify({message: msg})
				});
			}
		}
	};

	const event_emitter = eventEmitter(["message"]);

	instance.dispatchEvent = event_emitter.install(instance.public_interface);

	const event_source = SSEWithFetch(`${endpoint_url}/stream/${id}`);

	event_source.on("message", msg => {
		if (instance.is_ready === false) {
			if (msg === `C813A843-5765-4DB3-8439-24AC1849117D`) {
				instance.is_ready = true;

				ready_promise.resolve();

				setTimeout(() => {
					while (instance.message_queue.length) {
						const msg = instance.message_queue.shift();

						instance.dispatchEvent("message", msg);
					}
				}, 0);
			} else {
				instance.message_queue.push(JSON.parse(msg).message);
			}
		} else {
			instance.dispatchEvent("message", JSON.parse(msg).message);
		}
	});

	await ready_promise.promise;

	return instance.public_interface
}

export { index as createClient, index$1 as createServer };
